-- Deathstimator support

Deathstimator = {}
deathConf = {}
deathConf.show = nil
deathConf.inCombat = nil
Deathstimator.UnitHealthHistory = {}

SLASH_DEATHSTIMATOR1 = "/deathstimator"

function Deathstimator.OnLoad()
    this:RegisterEvent("PLAYER_LOGIN")
end

function Deathstimator.OnEvent()
    if event == "PLAYER_LOGIN" then
        deathConf.healthHistorySize = 10
        deathConf.minEntries = 6
        deathConf.registeredUnits = {
            "target",
            "player",
        }
        
        -- To initialize a fresh install
        if (deathConf.show == nil) or (deathConf.inCombat == nil) then
            deathConf.show = true
            deathConf.inCombat = false
        end

        -- To deal with kooky way previous version handled this
        if deathConf.inCombat and deathConf.show then
            deathConf.show = false
        end

        Deathstimator.inCombat = nil
        -- Show minion
        if deathConf.show then
            DeathstimatorMinionFrame:Show()
            Deathstimator.minion.SetText("Deathstimator")
        end

        -- Register units to track
        for idx, unitId in ipairs(deathConf.registeredUnits) do
            Deathstimator.UnitHealthHistory[unitId] = Deathstimator.HealthHistory:New()
        end

        this:RegisterEvent("UNIT_HEALTH")
        this:RegisterEvent("PLAYER_REGEN_ENABLED")
        this:RegisterEvent("PLAYER_REGEN_DISABLED")
        this:RegisterEvent("PLAYER_TARGET_CHANGED")

        SlashCmdList["DEATHSTIMATOR"] = Deathstimator.SlashCommand

    elseif event == "UNIT_HEALTH" then
        Deathstimator.OnUnitHealth(arg1)

    elseif event == "PLAYER_REGEN_ENABLED" then
        for _, historyObj in pairs(Deathstimator.UnitHealthHistory) do
            historyObj:Reset()
        end
        Deathstimator.inCombat = false
        Deathstimator.minion.SetText("Deathstimator")
        if deathConf.inCombat and DeathstimatorMinionFrame:IsShown() then
            DeathstimatorMinionFrame:Hide()
        end

    elseif event == "PLAYER_REGEN_DISABLED" then
        for _, historyObj in pairs(Deathstimator.UnitHealthHistory) do
            historyObj:Reset()
        end
        Deathstimator.inCombat = true
        if deathConf.inCombat and (not DeathstimatorMinionFrame:IsShown()) then
            DeathstimatorMinionFrame:Show()
        end

    elseif event == "PLAYER_TARGET_CHANGED" then
        Deathstimator.UnitHealthHistory["target"]:Reset()

    end
    
end

function Deathstimator.OnUnitHealth(unitId)
    for key, object in pairs(Deathstimator.UnitHealthHistory) do
        if (unitId == key) then
            -- Doesn't matter if it's hp or hpp
            local hp = UnitHealth(unitId)
            if not hp then
                return
            end
            -- Player and target have better ways of knowing when to reset.
            if unitId ~= "player" or unitId ~= "target" then
                -- Health changes are outside of combat, must be healing so reset.
                if not UnitAffectingCombat(unitId) then
                    object:Reset()
                    return
                end
            end
            object:AddInfo(hp)
            return
        end
    end
end


-- Target Health History Object
-- Used for computing the best fit line (slope) for the Target's health.
--
-- Compute the best slope using the method of least squares.
-- http://www.acad.sunytccc.edu/instruct/sbrown/stat/leastsq.htm
--
-- Wow, I NEVER thought I'd ever use ANYTHING like this in real life.
-- Haha, and then, it's only for a computer game.
-- Ms. Chen, my high school Calculus teacher, would be so proud.  Nah,
-- probably still too pissed at me.
--

Deathstimator.HealthHistory = {}

function Deathstimator.HealthHistory:New()
    local obj = {}
    setmetatable(obj, { __index = self })
    Deathstimator.HealthHistory.Reset(obj)
    return obj
end

function Deathstimator.HealthHistory:Reset()
    self.info = {}
    self.ctr = 0
    self.lastCtrComputed = 0
    self.m = 0
end

function Deathstimator.HealthHistory:AddInfo(hp)
    local now = GetTime()

    self.ctr = self.ctr + 1

    table.insert(self.info, { time = now, hp = hp })

    -- Remove entries which are older than the configured sample window,
    -- but retain a minimum number of entries.
    local cutOffTime = now - deathConf.healthHistorySize
    while table.getn(self.info) > deathConf.minEntries do
        if self.info[1].time >= cutOffTime then
            break
        end
        table.remove(self.info, 1)
    end
end

function Deathstimator.HealthHistory:GetN()
   return table.getn(self.info)
end

function Deathstimator.HealthHistory:ComputeSlope()
    local n = self:GetN()
    if (n == 0) then
        return nil
    end
    if (self.ctr == self.lastCtrComputed) then
        return self.m
    end
    self.lastCtrComputed = self.ctr

    local xSum = 0
    local ySum = 0
    local xSquaredSum = 0
    local xySum = 0
    for idx, healthInfo in ipairs(self.info) do
        local time = healthInfo.time
        local hp = healthInfo.hp
        -- time is the x access, hp is the y access
        xSum = xSum + time
        ySum = ySum + hp
        xSquaredSum = xSquaredSum + (time * time)
        xySum = xySum + (time * hp)
    end

    if (xSquaredSum == 0 or xSum == 0) then
        return nil
    end

    self.m = ( (n * xySum) - (xSum * ySum) ) / ( (n * xSquaredSum) - (xSum * xSum) )

    return self.m
end



-- Deathstimator minion
Deathstimator.minion = {}
Deathstimator.minion.lastUpdate = 0
Deathstimator.minion.updateInterval = 0.25
Deathstimator.minion.trackedUnit = nil

function Deathstimator.minion.OnUpdate()
    if not Deathstimator.minion.trackedUnit then
        unitId = "target"
    else
        unitId = Deathstimator.minion.trackedUnit
    end

    local now = GetTime()
    if (now >= (Deathstimator.minion.lastUpdate + Deathstimator.minion.updateInterval)) then
        Deathstimator.minion.lastUpdate = now

        if (not UnitExists(unitId)) then
            return
        end

        if Deathstimator.inCombat then
            local text
            local m, timeToDeath = Deathstimator.timeToDeath(unitId)
            if m == nil then
                text = "...gathering..."
            elseif m > 0 then
                -- Target is gaining health. Recalibrate.
                text = "...recalibrating..."
            else
                text = (UnitName(unitId) or "Unknown unit")..": Death in "..string.format("%.1f", timeToDeath).."s"
            end
            Deathstimator.minion.SetText(text)
        end
    end
end

function Deathstimator.minion.OnEnter(button)
   GameTooltip:SetOwner(button, "ANCHOR_RIGHT")
   GameTooltip:AddLine(" Deathstimator.\n")
   GameTooltip:AddLine("Shift + Left Click to move me around.\n")
   GameTooltip:Show()
end

function Deathstimator.minion.OnLeave(button)
   GameTooltip:Hide()
end

function Deathstimator.minion.SetText(text)
   if (not text) then
      text = ""
   end
   DeathstimatorMinionText:SetText(text)
end

function Deathstimator.timeToDeath(unitId)
    if (not UnitExists(unitId)) then
        return nil
    end

    local currentHp = UnitHealth(unitId)
    if (not currentHp or currentHp == 0) then
        return nil
    end

    local n = Deathstimator.UnitHealthHistory[unitId]:GetN()
    -- we want at least two data points to make a line
    if (n < 2) then
        return nil
    end

    local m = Deathstimator.UnitHealthHistory[unitId]:ComputeSlope()
    -- m is the slope, or hp(p) per second
    if (not m) then
        return nil
    end

    return m, math.abs(currentHp / m)
end



-- Deathstimator command line options
function Deathstimator.SlashCommand(line)
    local args = Deathstimator.SplitArgs(line)
    local cmd = args[1]

    if (not cmd or cmd == "" or cmd == "help") then
        Deathstimator.Help()

    elseif cmd == "show" then
        if deathConf.show then
            deathConf.show = false
            DeathstimatorMinionFrame:Hide()
            Deathstimator.chat("show: Never show Deathstimator minion")                        
        else
            deathConf.show = true
            if deathConf.inCombat then deathConf.inCombat = false end
            DeathstimatorMinionFrame:Show()
            Deathstimator.chat("show: Always show Deathstimator minion")            
        end
        Deathstimator.minion.SetText("Deathstimator")

    elseif cmd == "track" then
        local unitId = args[2]
        if not(unitId) then
            Deathstimator.chat("track: Invalid unitID: \""..tostring(unitId).."\"")
            return
        end
        Deathstimator.trackUnit(unitId)
        Deathstimator.trackUnitInMinion(unitId)
        Deathstimator.chat("track: Now tracking unit: \""..unitId.."\"")

    elseif cmd == "incombat" then
        if deathConf.inCombat then
            deathConf.inCombat = false
            Deathstimator.chat("incombat: Show minion in combat turned off")
            -- Hide minion
            Deathstimator.minion.SetText("Deathstimator")
            DeathstimatorMinionFrame:Hide()
        else
            deathConf.inCombat = true
            Deathstimator.chat("incombat: Show minion in combat turned on")
            -- Show minion
            if deathConf.show then deathConf.show = false end
            if Deathstimator.inCombat and (not DeathstimatorMinionFrame:IsShown()) then
                DeathstimatorMinionFrame:Show()
                Deathstimator.minion.SetText("Deathstimator")
            end
        end

    else
        Deathstimator.Help()

    end
end

function Deathstimator.Help()
    Deathstimator.chat("List of available command line options for the Deathstimator:")
    Deathstimator.chat("  help - Displays this message")
    Deathstimator.chat("  show - Always show minion ["..tostring(deathConf.show).."]")
    Deathstimator.chat("  incombat - Show minion in combat only ["..tostring(deathConf.inCombat).."]")
    Deathstimator.chat("  track <UnitId> - Track units other than \"target\"")
end

function Deathstimator.chat(msg)
   if (not r) then
      r = .9
      g = .9
      b = .9
   end
   DEFAULT_CHAT_FRAME:AddMessage(msg, r, g, b)
end

function Deathstimator.trackUnit(unitId)
    if not Deathstimator.UnitHealthHistory[unitId] then
        Deathstimator.UnitHealthHistory[unitId] = Deathstimator.HealthHistory:New()
    end

end

function Deathstimator.trackUnitInMinion(unitId)
    Deathstimator.minion.trackedUnit = unitId
end

function Deathstimator.split(s, delim)
   local pieces = {}
   local init = 1

   -- strip off any leading delimiters
   s = string.gsub(s, "^"..delim.."+", "")

   delim = "[^\\]"..delim

   while true do
      local start, send = string.find(s, delim, init)
      if (not start) then
         break
      end
      table.insert(pieces, string.sub(s, init, start))
      init = send + 1
   end

   table.insert(pieces, string.sub(s, init))

   return pieces
end

function Deathstimator.SplitArgs(line)
   return Deathstimator.split(line, "%s")
end

